# Title          : discretization.py
# Author         : Henry Lin <hlin117@gmail.com>
# License        : BSD 3 clause
#==============================================================================

from math import floor, log10

import numpy as np

from sklearn.base import BaseEstimator
from sklearn.base import TransformerMixin
from sklearn.utils import (
    check_array,
    check_X_y,
    column_or_1d,
    check_random_state,
)
from sklearn.utils.validation import check_is_fitted
from mdlp._mdlp import MDLPDiscretize


# def normalize(cut_points, _range, precision):
#     # if len(cut_points) == 0:
#         # return cut_points
#     # _range = np.max(col) - np.min(col)
#     multiplier = 10**(-floor(log10(_range))) / precision
#     return (cut_points * multiplier).astype(np.int) / multiplier


class MDLP(BaseEstimator, TransformerMixin):
    """Bins continuous values using MDLP "expert binning" method.

    Implements the MDLP discretization algorithm from Usama Fayyad's
    paper "Multi-Interval Discretization of Continuous-Valued
    Attributes for Classification Learning". Given the class labels
    for each sample, this transformer attempts to discretize a
    continuous attribute by minimizing the entropy at each interval.

    Parameters
    ----------
    continuous_features : 
        - None (default): All features are treated as continuous for discretization.
        - array of indices: Array of continous feature indices.
        - mask: Array of length n_features and with dtype=bool.

        If `X` is a 1-D array, then continuous_features is neglected.

    min_depth : int (default=0)
        The minimum depth of the interval splitting. Overrides
        the MDLP stopping criterion. If the entropy at a given interval
        is found to be zero before `min_depth`, the algorithm will stop.

    random_state : int (default=None)
        Seed of pseudo RNG to use when shuffling the data. Affects the
        outcome of MDLP if there are multiple samples with the same
        continuous value, but with different class labels.

    min_split : float (default=1e-3)
        The minmum size to split a bin

    dtype : np.dtype (default=np.int)
        The dtype of the transformed X

    Attributes
    ----------
    continuous_features_ : array-like of type int
        Similar to continous_features. However, for 2-D arrays, this
        attribute cannot be None.

    cut_points_ : dict of type {int : np.array}
        Dictionary mapping indices to a numpy array. Each
        numpy array is a sorted list of cut points found from
        discretization.

    dimensions_ : int
        Number of dimensions to input `X`. Either 1 or 2.

    Examples
    --------
    ```
        >>> from mdlp.discretization import MDLP
        >>> from sklearn.datasets import load_iris
        >>> iris = load_iris()
        >>> X = iris.data
        >>> y = iris.target
        >>> mdlp = MDLP()
        >>> conv_X = mdlp.fit_transform(X, y)

        `conv_X` will be the same shape as `X`, except it will contain
        integers instead of continuous attributes representing the results
        of the discretization process.

        To retrieve the explicit intervals of the discretization of, say,
        the third column (index 2), one can do

        >>> mdlp.cat2intervals(conv_X, 2)

        which would return a list of tuples `(a, b)`. Each tuple represents
        the contnuous interval (a, b], where `a` can be `float("-inf")`,
        and `b` can be `float("inf")`.
    ```
    """

    def __init__(self, continuous_features=None, min_depth=0, random_state=None, 
                 min_split=1e-3, dtype=np.int):
        # Parameters
        # self.continuous_features = None
        self.min_depth = min_depth
        self.random_state = random_state
        self.min_split = min_split
        self.continuous_features = continuous_features
        self.dtype = dtype

        # Attributes
        self.continuous_features_ = None
        self.cut_points_ = None
        self.mins_ = None
        self.maxs_ = None

    def fit(self, X, y):
        """Finds the intervals of interest from the input data.

        Parameters
        ----------
        X : The array containing features to be discretized. Continuous
            features should be specified by the `continuous_features`
            attribute if `X` is a 2-D array.

        y : A list or array of class labels corresponding to `X`.

        continuous_features : (default None) a list of indices that you want to discretize
                              or a list (or array) of bools indicating the continuous features
        """
        X = check_array(X, force_all_finite=True, ensure_2d=True, dtype=np.float64)
        y = column_or_1d(y)
        y = check_array(y, ensure_2d=False, dtype=np.int)
        X, y = check_X_y(X, y)

        if len(X.shape) != 2:
            raise ValueError("Invalid input dimension for `X`. "
                             "Input shape is expected to be 2D, but is {0}".format(X.shape))

        state = check_random_state(self.random_state)
        perm = state.permutation(len(y))
        X = X[perm]
        y = y[perm]

        if self.continuous_features is None:
            self.continuous_features_ = np.arange(X.shape[1])
        else:
            continuous_features = np.array(self.continuous_features)
            if continuous_features.dtype == np.bool:
                continuous_features = np.arange(len(continuous_features))[continuous_features]
            else:
                continuous_features = continuous_features.astype(np.int, casting='safe')
                assert np.max(continuous_features) < X.shape[1] and np.min(continuous_features) >= 0
            self.continuous_features_ = continuous_features

        self.cut_points_ = [None] * X.shape[1]
        mins = np.min(X, axis=0)
        maxs = np.max(X, axis=0)

        for index in self.continuous_features_:
            col = X[:, index]
            cut_points = MDLPDiscretize(col, y, self.min_depth, self.min_split)
            self.cut_points_[index] = cut_points
            # self.cut_points_[index] = normalize(cut_points, maxs[index] - mins[index], self.precision)

        self.mins_ = mins
        self.maxs_ = maxs
        return self

    def transform(self, X):
        """Discretizes values in X into {0, ..., k-1}.

        `k` is the number of bins the discretizer creates from a continuous
        feature.
        """
        X = check_array(X, force_all_finite=True, ensure_2d=False)
        check_is_fitted(self, "cut_points_")

        output = X.copy()
        for i in self.continuous_features_:
            output[:, i] = np.searchsorted(self.cut_points_[i], X[:, i])
        return output.astype(self.dtype)

    def cat2intervals(self, X, index):
        """Converts categorical data into intervals.

        Parameters
        ----------
        X : The discretized array

        index: which feature index to convert
        """

        cp_indices = X[:, index]
        return self.assign_intervals(cp_indices, index)

    def cts2cat(self, col, index):
        """Converts each continuous feature from index `index` into
        a categorical feature from the input column `col`.
        """
        return np.searchsorted(self.cut_points_[index], col)

    def assign_intervals(self, cp_indices, index):
        """Assigns the cut point indices `cp_indices` (representing
        categorical features) into a list of intervals.
        """

        # Case for a 1-D array
        cut_points = self.cut_points_[index]
        if cut_points is None:
            raise ValueError("The given index %d has not been discretized!")
        non_zero_mask = cp_indices[cp_indices - 1 != -1].astype(int) - 1
        fronts = np.zeros(cp_indices.shape)
        fronts[cp_indices == 0] = float("-inf")
        fronts[cp_indices != 0] = cut_points[non_zero_mask]

        n_cuts = len(cut_points)
        backs = np.zeros(cp_indices.shape)
        non_n_cuts_mask = cp_indices[cp_indices != n_cuts].astype(int)
        backs[cp_indices == n_cuts] = float("inf")
        backs[cp_indices != n_cuts] = cut_points[non_n_cuts_mask]

        return [(front, back) for front, back in zip(fronts, backs)]

    # def



#!/usr/bin/python
"""
Implements the MDLP discretization criterion from Usama Fayyad's paper
"Multi-Interval Discretization of Continuous-Valued Attributes for
Classification Learning."
"""

from setuptools import Extension, find_packages, setup

if __name__ == '__main__':
    # see https://stackoverflow.com/a/42163080 for the approach to pushing
    # numpy and cython dependencies into extension building only
    try:
        # if Cython is available, we will rebuild from the pyx file directly
        from Cython.Distutils import build_ext
        sources = ['mdlp/_mdlp.pyx']
    except:
        # else we build from the cpp file included in the distribution
        from setuptools.command.build_ext import build_ext
        sources = ['mdlp/_mdlp.cpp']

    class CustomBuildExt(build_ext):
        """Custom build_ext class to defer numpy imports until needed.

        Overrides the run command for building an extension and adds in numpy
        include dirs to the extension build. Doing this at extension build time
        allows us to avoid requiring that numpy be pre-installed before
        executing this setup script.
        """

        def run(self):
            import numpy
            self.include_dirs.append(numpy.get_include())
            build_ext.run(self)

    cpp_ext = Extension(
        'mdlp._mdlp',
        sources=sources,
        libraries=[],
        include_dirs=[],
        language='c++',
    )

    setup(
        name='mdlp-discretization',
        version='0.3.3rc',
        description=__doc__,
        license='BSD 3 Clause',
        url='github.com/hlin117/mdlp-discretization',
        author='Henry Lin',
        author_email='hlin117@gmail.com',
        install_requires=[
            # 'setuptools>=18.0',
            'numpy>=1.11.2',
            'scipy>=0.18.1',
            'scikit-learn>=0.18.1',
            # 'pytest>=3.2.2',
        ],
        packages=find_packages(),
        ext_modules=[cpp_ext],
        cmdclass={'build_ext': CustomBuildExt},
    )



from __future__ import division
import cython
import numpy as np
cimport numpy as np

from cpython.mem cimport PyMem_Free
from cpython.mem cimport PyMem_Malloc
from libc.math cimport INFINITY
from libc.math cimport log, pow
from libcpp.set cimport set as stdset
from libcpp.vector cimport vector as stdvector
from scipy.stats import entropy

# some basic type definitions
ctypedef np.npy_intp SIZE_t  # Type for indices and counters
ctypedef SIZE_t* LEVEL
ctypedef np.float64_t FLOAT
cdef SIZE_t LEVEL_SIZE = sizeof(SIZE_t) * 3
cdef inline void set_level(LEVEL level, SIZE_t start, SIZE_t end, SIZE_t depth):
    level[0] = start
    level[1] = end
    level[2] = depth


@cython.boundscheck(False)
def MDLPDiscretize(col, y, int min_depth, FLOAT min_split):
    """Performs MDLP discretization on X and y"""

    order = np.argsort(col)
    col = col[order]
    y = y[order]

    cdef stdset[FLOAT] cut_points = stdset[FLOAT]()

    # Now we do a depth first search to create cut_points
    cdef int num_samples = len(col)
    cdef int min_span = (int) (min_split * num_samples)
    cdef LEVEL init_level = <LEVEL> PyMem_Malloc(LEVEL_SIZE)
    cdef stdvector[LEVEL] search_intervals = stdvector[LEVEL]()
    set_level(init_level, 0, num_samples, 0)

    search_intervals.push_back(init_level)
    cdef LEVEL currlevel
    while search_intervals.size() > 0:
        currlevel = search_intervals.back()
        search_intervals.pop_back()
        start, end, depth = unwrap(currlevel)
        PyMem_Free(currlevel)
        if end - start <= min_split:
            break
        k = find_cut(y, start, end)

        # Need to see whether the "front" and "back" of the interval need
        # to be float("-inf") or float("inf")
        if (k == -1) or (depth >= min_depth and reject_split(y, start, end, k)):
            front = -1 * INFINITY if (start == 0) else get_cut(col, start)
            back = INFINITY if (end == num_samples) else get_cut(col, end)

            if front == back: continue  # Corner case
            if front != -1 * INFINITY:
                cut_points.insert(front)
            if back != INFINITY:
                cut_points.insert(back)

            continue

        left_level = <LEVEL> PyMem_Malloc(LEVEL_SIZE)
        right_level = <LEVEL> PyMem_Malloc(LEVEL_SIZE)
        set_level(left_level, start, k, depth+1)
        set_level(right_level, k, end, depth+1)
        search_intervals.push_back(left_level)
        search_intervals.push_back(right_level)

    output = np.array([num for num in cut_points])
    output = np.sort(output)
    return output

cdef unwrap(LEVEL level):
    return level[0], level[1], level[2]

@cython.boundscheck(False)
cdef float get_cut(np.ndarray[np.float64_t, ndim=1] col, int ind):
    return (col[ind-1] + col[ind]) / 2

@cython.boundscheck(False)
def slice_entropy(np.ndarray[np.int_t, ndim=1] y, SIZE_t start, SIZE_t end):
    """Returns the entropy of the given slice of y. Also returns the
    number of classes within the interval.

    NOTE: The entropy function scipy provides uses the natural log,
    not log base 2.
    """
    counts = np.bincount(y[start:end])
    vals = np.true_divide(counts, end - start)
    return entropy(vals), np.sum(vals != 0)

@cython.boundscheck(False)
cdef bint reject_split(np.ndarray[np.int_t, ndim=1] y, int start, int end, int k):
    """Using the minimum description length principal, determines
    whether it is appropriate to stop cutting.
    """

    cdef float N = <float> (end - start)
    entropy1, k1 = slice_entropy(y, start, k)
    entropy2, k2 = slice_entropy(y, k, end)
    whole_entropy, k0 = slice_entropy(y, start, end)

    # Calculate the final values
    cdef:
        float part1 = 1 / N * ((k - start) * entropy1 + (end - k) * entropy2)
        float gain = whole_entropy - part1
        float entropy_diff = k0 * whole_entropy - k1 * entropy1 - k2 * entropy2
        float delta = log(pow(3, k0) - 2) - entropy_diff
    return gain <= 1 / N * (log(N - 1) + delta)

@cython.boundscheck(False)
def find_cut(np.ndarray[np.int_t, ndim=1] y, int start, int end):
    """Finds the cut point between `start` and `end`.

    Returns an index, `k`, which splits `y` into two sub arrays A and B,
    where `k` is the first index of B. Thus, if `n` is the length of
    `y`, then `k` is in the range {1, 2, ..., n}. Visually, `k` may imply
    the following sub arrays A and B:

        [0, 1, 2, 3, ..., k - 1] [k, k + 2, ..., n]

    If k = -1, then there is no index `k` which satisfies the MDLP cutting
    criterion.

    Parameters
    ----------
    y : 1-d numpy array
        class labels

    start : int
        From where we start probing `y` for the best cut.

    end : int
        End of where we start probing `y` for the best cut. Index
        is exclusive.

    Returns
    -------
    k : int
        Integer index of the best cut.
    """

    cdef:
        int length = end - start
        float prev_entropy = np.inf  #INFINITY
        SIZE_t k = -1
        int ind
        float first_half, second_half, curr_entropy
    for ind in range(start + 1, end):

        # I choose not to use a `min` function here for this optimization.
        if y[ind-1] == y[ind]:
            continue

        # Find the partition entropy, and check if this entropy is minimum
        first_half = (<float> (ind - start)) / length \
                            * slice_entropy(y, start, ind)[0]
        second_half = (<float> (end - ind)) / length \
                             * slice_entropy(y, ind, end)[0]
        curr_entropy = first_half + second_half

        if prev_entropy > curr_entropy:
            prev_entropy = curr_entropy
            k = ind

    return k  # NOTE: k == -1 if there is no good cut
